@@ -20,6 +20,7 @@
 
 */
 
+#define PY_SSIZE_T_CLEAN
 #include <Python.h>
 
 #include <sys/types.h>
@@ -32,33 +33,6 @@
 #define get_perm acl_get_perm_np
 #endif
 
-#if PY_MAJOR_VERSION >= 3
-#define IS_PY3K
-#define PyInt_Check(op) PyLong_Check(op)
-#define PyInt_FromString PyLong_FromString
-#define PyInt_FromUnicode PyLong_FromUnicode
-#define PyInt_FromLong PyLong_FromLong
-#define PyInt_FromSize_t PyLong_FromSize_t
-#define PyInt_FromSsize_t PyLong_FromSsize_t
-#define PyInt_AsLong PyLong_AsLong
-#define PyInt_AsSsize_t PyLong_AsSsize_t
-#define PyInt_AsUnsignedLongMask PyLong_AsUnsignedLongMask
-#define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
-#define PyInt_AS_LONG PyLong_AS_LONG
-#define MyString_FromFormat PyUnicode_FromFormat
-#define MyString_FromString PyUnicode_FromString
-#define MyString_FromStringAndSize PyUnicode_FromStringAndSize
-#else
-#define PyBytes_Check PyString_Check
-#define PyBytes_AS_STRING PyString_AS_STRING
-#define PyBytes_FromStringAndSize PyString_FromStringAndSize
-#define PyBytes_FromString PyString_FromString
-#define PyBytes_FromFormat PyString_FromFormat
-#define MyString_FromFormat PyBytes_FromFormat
-#define MyString_FromString PyBytes_FromString
-#define MyString_FromStringAndSize PyBytes_FromStringAndSize
-#endif
-
 /* Used for cpychecker: */
 /* The checker automatically defines this preprocessor name when creating
    the custom attribute: */
@@ -131,15 +105,24 @@ typedef struct {
 static PyObject* ACL_new(PyTypeObject* type, PyObject* args,
                          PyObject *keywds) {
     PyObject* newacl;
+    ACL_Object *acl;
 
     newacl = type->tp_alloc(type, 0);
 
-    if(newacl != NULL) {
-        ((ACL_Object*)newacl)->acl = NULL;
-#ifdef HAVEL_LEVEL2
-        ((ACL_Object*)newacl)->entry_id = ACL_FIRST_ENTRY;
-#endif
+    if(newacl == NULL) {
+        return NULL;
     }
+    acl = (ACL_Object*) newacl;
+
+    acl->acl = acl_init(0);
+    if (acl->acl == NULL) {
+        PyErr_SetFromErrno(PyExc_IOError);
+        Py_DECREF(newacl);
+        return NULL;
+    }
+#ifdef HAVE_LEVEL2
+    acl->entry_id = ACL_FIRST_ENTRY;
+#endif
 
     return newacl;
 }
@@ -147,20 +130,36 @@ static PyObject* ACL_new(PyTypeObject* type, PyObject* args,
 /* Initialization of a new ACL instance */
 static int ACL_init(PyObject* obj, PyObject* args, PyObject *keywds) {
     ACL_Object* self = (ACL_Object*) obj;
-#ifdef HAVE_LINUX
     static char *kwlist[] = { "file", "fd", "text", "acl", "filedef",
-                              "mode", NULL };
-    char *format = "|etisO!si";
-    int mode = -1;
-#else
-    static char *kwlist[] = { "file", "fd", "text", "acl", "filedef", NULL };
-    char *format = "|etisO!s";
+#ifdef HAVE_LINUX
+                              "mode",
 #endif
-    char *file = NULL;
-    char *filedef = NULL;
+#ifdef HAVE_ACL_COPY_EXT
+                              "data",
+#endif
+                              NULL };
+    char *format = "|O&OsO!O&"
+#ifdef HAVE_LINUX
+      "i"
+#endif
+#ifdef HAVE_ACL_COPY_EXT
+      "y#"
+#endif
+      ;
+    acl_t new = NULL;
+#ifdef HAVE_LINUX
+    int mode = -1;
+#endif
+    PyObject *file = NULL;
+    PyObject *filedef = NULL;
     char *text = NULL;
-    int fd = -1;
+    PyObject *fd = NULL;
     ACL_Object* thesrc = NULL;
+#ifdef HAVE_ACL_COPY_EXT
+    const void *buf = NULL;
+    Py_ssize_t bufsize;
+#endif
+    int set_err = 0;
 
     if(!PyTuple_Check(args) || PyTuple_Size(args) != 0 ||
        (keywds != NULL && PyDict_Check(keywds) && PyDict_Size(keywds) > 1)) {
@@ -169,40 +168,73 @@ static int ACL_init(PyObject* obj, PyObject* args, PyObject *keywds) {
         return -1;
     }
     if(!PyArg_ParseTupleAndKeywords(args, keywds, format, kwlist,
-                                    NULL, &file, &fd, &text, &ACL_Type,
-                                    &thesrc, &filedef
+                                    PyUnicode_FSConverter, &file,
+                                    &fd, &text, &ACL_Type, &thesrc,
+                                    PyUnicode_FSConverter, &filedef
 #ifdef HAVE_LINUX
                                     , &mode
+#endif
+#ifdef HAVE_ACL_COPY_EXT
+                                    , &buf, &bufsize
 #endif
                                     ))
         return -1;
 
+    if(file != NULL) {
+        char *path = PyBytes_AS_STRING(file);
+        new = acl_get_file(path, ACL_TYPE_ACCESS);
+        // Set custom exception on this failure path which includes
+        // the filename.
+        if (new == NULL) {
+          PyErr_SetFromErrnoWithFilename(PyExc_IOError, path);
+          set_err = 1;
+        }
+        Py_DECREF(file);
+    } else if(text != NULL)
+        new = acl_from_text(text);
+    else if(fd != NULL) {
+        int fdval;
+        if ((fdval = PyObject_AsFileDescriptor(fd)) != -1) {
+            new = acl_get_fd(fdval);
+        }
+    } else if(thesrc != NULL)
+        new = acl_dup(thesrc->acl);
+    else if(filedef != NULL) {
+        char *path = PyBytes_AS_STRING(filedef);
+        new = acl_get_file(path, ACL_TYPE_DEFAULT);
+        // Set custom exception on this failure path which includes
+        // the filename.
+        if (new == NULL) {
+          PyErr_SetFromErrnoWithFilename(PyExc_IOError, path);
+          set_err = 1;
+        }
+        Py_DECREF(filedef);
+    }
+#ifdef HAVE_LINUX
+    else if(mode != -1)
+        new = acl_from_mode(mode);
+#endif
+#ifdef HAVE_ACL_COPY_EXT
+    else if(buf != NULL) {
+      new = acl_copy_int(buf);
+    }
+#endif
+    else
+        new = acl_init(0);
+
+    if(new == NULL) {
+        if (!set_err) {
+            PyErr_SetFromErrno(PyExc_IOError);
+        }
+        return -1;
+    }
+
     /* Free the old acl_t without checking for error, we don't
      * care right now */
     if(self->acl != NULL)
         acl_free(self->acl);
 
-    if(file != NULL)
-        self->acl = acl_get_file(file, ACL_TYPE_ACCESS);
-    else if(text != NULL)
-        self->acl = acl_from_text(text);
-    else if(fd != -1)
-        self->acl = acl_get_fd(fd);
-    else if(thesrc != NULL)
-        self->acl = acl_dup(thesrc->acl);
-    else if(filedef != NULL)
-        self->acl = acl_get_file(filedef, ACL_TYPE_DEFAULT);
-#ifdef HAVE_LINUX
-    else if(mode != -1)
-        self->acl = acl_from_mode(mode);
-#endif
-    else
-        self->acl = acl_init(0);
-
-    if(self->acl == NULL) {
-        PyErr_SetFromErrno(PyExc_IOError);
-        return -1;
-    }
+    self->acl = new;
 
     return 0;
 }
@@ -232,7 +264,7 @@ static PyObject* ACL_str(PyObject *obj) {
     if(text == NULL) {
         return PyErr_SetFromErrno(PyExc_IOError);
     }
-    ret = MyString_FromString(text);
+    ret = PyUnicode_FromString(text);
     if(acl_free(text) != 0) {
         Py_XDECREF(ret);
         return PyErr_SetFromErrno(PyExc_IOError);
@@ -384,16 +416,7 @@ static PyObject* ACL_equiv_mode(PyObject* obj, PyObject* args) {
 
     if(acl_equiv_mode(self->acl, &mode) == -1)
         return PyErr_SetFromErrno(PyExc_IOError);
-    return PyInt_FromLong(mode);
-}
-#endif
-
-#ifndef IS_PY3K
-/* Implementation of the compare for ACLs */
-static int ACL_nocmp(PyObject* o1, PyObject* o2) {
-
-    PyErr_SetString(PyExc_TypeError, "cannot compare ACLs using cmp()");
-    return -1;
+    return PyLong_FromLong(mode);
 }
 #endif
 
@@ -411,40 +434,38 @@ static char __applyto_doc__[] =
 /* Applies the ACL to a file */
 static PyObject* ACL_applyto(PyObject* obj, PyObject* args) {
     ACL_Object *self = (ACL_Object*) obj;
-    PyObject *myarg;
+    PyObject *target, *tmp;
     acl_type_t type = ACL_TYPE_ACCESS;
     int nret;
     int fd;
 
-    if (!PyArg_ParseTuple(args, "O|I", &myarg, &type))
+    if (!PyArg_ParseTuple(args, "O|I", &target, &type))
         return NULL;
-
-    if(PyBytes_Check(myarg)) {
-        char *filename = PyBytes_AS_STRING(myarg);
-        nret = acl_set_file(filename, type, self->acl);
-    } else if (PyUnicode_Check(myarg)) {
-        PyObject *o =
-            PyUnicode_AsEncodedString(myarg,
-                                      Py_FileSystemDefaultEncoding, "strict");
-        if (o == NULL)
-            return NULL;
-        const char *filename = PyBytes_AS_STRING(o);
-        nret = acl_set_file(filename, type, self->acl);
-        Py_DECREF(o);
-    } else if((fd = PyObject_AsFileDescriptor(myarg)) != -1) {
-        nret = acl_set_fd(fd, self->acl);
+    if ((fd = PyObject_AsFileDescriptor(target)) != -1) {
+        if((nret = acl_set_fd(fd, self->acl)) == -1) {
+          PyErr_SetFromErrno(PyExc_IOError);
+        }
     } else {
-        PyErr_SetString(PyExc_TypeError, "argument 1 must be string, int,"
-                        " or file-like object");
-        return 0;
+      // PyObject_AsFileDescriptor sets an error when failing, so clear
+      // it such that further code works; some method lookups fail if an
+      // error already occured when called, which breaks at least
+      // PyOS_FSPath (called by FSConverter).
+      PyErr_Clear();
+      if(PyUnicode_FSConverter(target, &tmp)) {
+        char *filename = PyBytes_AS_STRING(tmp);
+        if ((nret = acl_set_file(filename, type, self->acl)) == -1) {
+            PyErr_SetFromErrnoWithFilename(PyExc_IOError, filename);
+        }
+        Py_DECREF(tmp);
+      } else {
+        nret = -1;
+      }
     }
-    if(nret == -1) {
-        return PyErr_SetFromErrno(PyExc_IOError);
+    if (nret < 0) {
+        return NULL;
+    } else {
+        Py_RETURN_NONE;
     }
-
-    /* Return the result */
-    Py_INCREF(Py_None);
-    return Py_None;
 }
 
 static char __valid_doc__[] =
@@ -516,11 +537,11 @@ static PyObject* ACL_get_state(PyObject *obj, PyObject* args) {
 static PyObject* ACL_set_state(PyObject *obj, PyObject* args) {
     ACL_Object *self = (ACL_Object*) obj;
     const void *buf;
-    int bufsize;
+    Py_ssize_t bufsize;
     acl_t ptr;
 
     /* Parse the argument */
-    if (!PyArg_ParseTuple(args, "s#", &buf, &bufsize))
+    if (!PyArg_ParseTuple(args, "y#", &buf, &bufsize))
         return NULL;
 
     /* Try to import the external representation */
@@ -535,9 +556,7 @@ static PyObject* ACL_set_state(PyObject *obj, PyObject* args) {
 
     self->acl = ptr;
 
-    /* Return the result */
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 #endif
 
@@ -601,12 +620,15 @@ static PyObject* ACL_delete_entry(PyObject *obj, PyObject *args) {
     if (!PyArg_ParseTuple(args, "O!", &Entry_Type, &e))
         return NULL;
 
+    if (e->parent_acl != obj) {
+        PyErr_SetString(PyExc_ValueError,
+                        "Can't remove un-owned entry");
+        return NULL;
+    }
     if(acl_delete_entry(self->acl, e->entry) == -1)
         return PyErr_SetFromErrno(PyExc_IOError);
 
-    /* Return the result */
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static char __ACL_calc_mask_doc__[] =
@@ -633,9 +655,7 @@ static PyObject* ACL_calc_mask(PyObject *obj, PyObject *args) {
     if(acl_calc_mask(&self->acl) == -1)
         return PyErr_SetFromErrno(PyExc_IOError);
 
-    /* Return the result */
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static char __ACL_append_doc__[] =
@@ -654,27 +674,24 @@ static char __ACL_append_doc__[] =
 
 /* Convenience method to create a new Entry */
 static PyObject* ACL_append(PyObject *obj, PyObject *args) {
-    ACL_Object* self = (ACL_Object*) obj;
     Entry_Object* newentry;
     Entry_Object* oldentry = NULL;
     int nret;
 
-    newentry = (Entry_Object*)PyType_GenericNew(&Entry_Type, NULL, NULL);
+    if (!PyArg_ParseTuple(args, "|O!", &Entry_Type, &oldentry)) {
+        return NULL;
+    }
+
+    PyObject *new_arglist = Py_BuildValue("(O)", obj);
+    if (new_arglist == NULL) {
+        return NULL;
+    }
+    newentry = (Entry_Object*) PyObject_CallObject((PyObject*)&Entry_Type, new_arglist);
+    Py_DECREF(new_arglist);
     if(newentry == NULL) {
         return NULL;
     }
 
-    if (!PyArg_ParseTuple(args, "|O!", &Entry_Type, &oldentry)) {
-        Py_DECREF(newentry);
-        return NULL;
-    }
-
-    nret = acl_create_entry(&self->acl, &newentry->entry);
-    if(nret == -1) {
-        Py_DECREF(newentry);
-        return PyErr_SetFromErrno(PyExc_IOError);
-    }
-
     if(oldentry != NULL) {
         nret = acl_copy_entry(newentry->entry, oldentry->entry);
         if(nret == -1) {
@@ -683,9 +700,6 @@ static PyObject* ACL_append(PyObject *obj, PyObject *args) {
         }
     }
 
-    newentry->parent_acl = obj;
-    Py_INCREF(obj);
-
     return (PyObject*)newentry;
 }
 
@@ -735,18 +749,38 @@ static int get_tag_qualifier(acl_entry_t entry, tag_qual *tq) {
     return 0;
 }
 
+#define ENTRY_SET_CHECK(self, attr, value)         \
+    if (value == NULL) { \
+        PyErr_SetString(PyExc_TypeError, \
+                        attr " deletion is not supported"); \
+        return -1; \
+    }
+
 /* Creation of a new Entry instance */
 static PyObject* Entry_new(PyTypeObject* type, PyObject* args,
                            PyObject *keywds) {
     PyObject* newentry;
+    Entry_Object* entry;
+    ACL_Object* parent = NULL;
+
+    if (!PyArg_ParseTuple(args, "O!", &ACL_Type, &parent))
+        return NULL;
 
     newentry = PyType_GenericNew(type, args, keywds);
 
-    if(newentry != NULL) {
-        ((Entry_Object*)newentry)->entry = NULL;
-        ((Entry_Object*)newentry)->parent_acl = NULL;
+    if(newentry == NULL) {
+        return NULL;
     }
 
+    entry = (Entry_Object*)newentry;
+
+    if(acl_create_entry(&parent->acl, &entry->entry) == -1) {
+        PyErr_SetFromErrno(PyExc_IOError);
+        Py_DECREF(newentry);
+        return NULL;
+    }
+    Py_INCREF(parent);
+    entry->parent_acl = (PyObject*)parent;
     return newentry;
 }
 
@@ -758,14 +792,11 @@ static int Entry_init(PyObject* obj, PyObject* args, PyObject *keywds) {
     if (!PyArg_ParseTuple(args, "O!", &ACL_Type, &parent))
         return -1;
 
-    if(acl_create_entry(&parent->acl, &self->entry) == -1) {
-        PyErr_SetFromErrno(PyExc_IOError);
+    if ((PyObject*)parent != self->parent_acl) {
+        PyErr_SetString(PyExc_ValueError,
+                        "Can't reinitialize with a different parent");
         return -1;
     }
-
-    self->parent_acl = (PyObject*)parent;
-    Py_INCREF(parent);
-
     return 0;
 }
 
@@ -796,70 +827,61 @@ static PyObject* Entry_str(PyObject *obj) {
         return NULL;
     }
 
-    format = MyString_FromString("ACL entry for ");
+    format = PyUnicode_FromString("ACL entry for ");
     if(format == NULL)
         return NULL;
     switch(tq.tag) {
     case ACL_UNDEFINED_TAG:
-        kind = MyString_FromString("undefined type");
+        kind = PyUnicode_FromString("undefined type");
         break;
     case ACL_USER_OBJ:
-        kind = MyString_FromString("the owner");
+        kind = PyUnicode_FromString("the owner");
         break;
     case ACL_GROUP_OBJ:
-        kind = MyString_FromString("the group");
+        kind = PyUnicode_FromString("the group");
         break;
     case ACL_OTHER:
-        kind = MyString_FromString("the others");
+        kind = PyUnicode_FromString("the others");
         break;
     case ACL_USER:
         /* FIXME: here and in the group case, we're formatting with
            unsigned, because there's no way to automatically determine
            the signed-ness of the types; on Linux(glibc) they're
            unsigned, so we'll go along with that */
-        kind = MyString_FromFormat("user with uid %u", tq.uid);
+        kind = PyUnicode_FromFormat("user with uid %u", tq.uid);
         break;
     case ACL_GROUP:
-        kind = MyString_FromFormat("group with gid %u", tq.gid);
+        kind = PyUnicode_FromFormat("group with gid %u", tq.gid);
         break;
     case ACL_MASK:
-        kind = MyString_FromString("the mask");
+        kind = PyUnicode_FromString("the mask");
         break;
     default:
-        kind = MyString_FromString("UNKNOWN_TAG_TYPE!");
+        kind = PyUnicode_FromString("UNKNOWN_TAG_TYPE!");
         break;
     }
     if (kind == NULL) {
         Py_DECREF(format);
         return NULL;
     }
-#ifdef IS_PY3K
     PyObject *ret = PyUnicode_Concat(format, kind);
     Py_DECREF(format);
     Py_DECREF(kind);
     return ret;
-#else
-    PyString_ConcatAndDel(&format, kind);
-    return format;
-#endif
 }
 
 /* Sets the tag type of the entry */
 static int Entry_set_tag_type(PyObject* obj, PyObject* value, void* arg) {
     Entry_Object *self = (Entry_Object*) obj;
 
-    if(value == NULL) {
-        PyErr_SetString(PyExc_TypeError,
-                        "tag type deletion is not supported");
-        return -1;
-    }
+    ENTRY_SET_CHECK(self, "tag type", value);
 
-    if(!PyInt_Check(value)) {
+    if(!PyLong_Check(value)) {
         PyErr_SetString(PyExc_TypeError,
                         "tag type must be integer");
         return -1;
     }
-    if(acl_set_tag_type(self->entry, (acl_tag_t)PyInt_AsLong(value)) == -1) {
+    if(acl_set_tag_type(self->entry, (acl_tag_t)PyLong_AsLong(value)) == -1) {
         PyErr_SetFromErrno(PyExc_IOError);
         return -1;
     }
@@ -872,16 +894,12 @@ static PyObject* Entry_get_tag_type(PyObject *obj, void* arg) {
     Entry_Object *self = (Entry_Object*) obj;
     acl_tag_t value;
 
-    if (self->entry == NULL) {
-        PyErr_SetString(PyExc_AttributeError, "entry attribute");
-        return NULL;
-    }
     if(acl_get_tag_type(self->entry, &value) == -1) {
         PyErr_SetFromErrno(PyExc_IOError);
         return NULL;
     }
 
-    return PyInt_FromLong(value);
+    return PyLong_FromLong(value);
 }
 
 /* Sets the qualifier (either uid_t or gid_t) for the entry,
@@ -889,24 +907,23 @@ static PyObject* Entry_get_tag_type(PyObject *obj, void* arg) {
  */
 static int Entry_set_qualifier(PyObject* obj, PyObject* value, void* arg) {
     Entry_Object *self = (Entry_Object*) obj;
-    long uidgid;
+    unsigned long uidgid;
     uid_t uid;
     gid_t gid;
     void *p;
     acl_tag_t tag;
 
-    if(value == NULL) {
-        PyErr_SetString(PyExc_TypeError,
-                        "qualifier deletion is not supported");
-        return -1;
-    }
+    ENTRY_SET_CHECK(self, "qualifier", value);
 
-    if(!PyInt_Check(value)) {
+    if(!PyLong_Check(value)) {
         PyErr_SetString(PyExc_TypeError,
                         "qualifier must be integer");
         return -1;
     }
-    if((uidgid = PyInt_AsLong(value)) == -1) {
+    /* This is the negative value check, and larger than long
+       check. If uid_t/gid_t are long-sized, this is enough to check
+       for both over and underflow. */
+    if((uidgid = PyLong_AsUnsignedLong(value)) == (unsigned long) -1) {
         if(PyErr_Occurred() != NULL) {
             return -1;
         }
@@ -920,18 +937,20 @@ static int Entry_set_qualifier(PyObject* obj, PyObject* value, void* arg) {
     }
     uid = uidgid;
     gid = uidgid;
+    /* This is an extra overflow check, in case uid_t/gid_t are
+       int-sized (and int size smaller than long size). */
     switch(tag) {
     case ACL_USER:
-      if((long)uid != uidgid) {
-        PyErr_SetString(PyExc_OverflowError, "cannot assign given qualifier");
+      if((unsigned long)uid != uidgid) {
+        PyErr_SetString(PyExc_OverflowError, "Can't assign given qualifier");
         return -1;
       } else {
         p = &uid;
       }
       break;
     case ACL_GROUP:
-      if((long)gid != uidgid) {
-        PyErr_SetString(PyExc_OverflowError, "cannot assign given qualifier");
+      if((unsigned long)gid != uidgid) {
+        PyErr_SetString(PyExc_OverflowError, "Can't assign given qualifier");
         return -1;
       } else {
         p = &gid;
@@ -939,7 +958,7 @@ static int Entry_set_qualifier(PyObject* obj, PyObject* value, void* arg) {
       break;
     default:
       PyErr_SetString(PyExc_TypeError,
-                      "can only set qualifiers on ACL_USER or ACL_GROUP entries");
+                      "Can only set qualifiers on ACL_USER or ACL_GROUP entries");
       return -1;
     }
     if(acl_set_qualifier(self->entry, p) == -1) {
@@ -953,13 +972,9 @@ static int Entry_set_qualifier(PyObject* obj, PyObject* value, void* arg) {
 /* Returns the qualifier of the entry */
 static PyObject* Entry_get_qualifier(PyObject *obj, void* arg) {
     Entry_Object *self = (Entry_Object*) obj;
-    long value;
+    unsigned long value;
     tag_qual tq;
 
-    if (self->entry == NULL) {
-        PyErr_SetString(PyExc_AttributeError, "entry attribute");
-        return NULL;
-    }
     if(get_tag_qualifier(self->entry, &tq) < 0) {
         return NULL;
     }
@@ -969,11 +984,11 @@ static PyObject* Entry_get_qualifier(PyObject *obj, void* arg) {
         value = tq.gid;
     } else {
         PyErr_SetString(PyExc_TypeError,
-                        "given entry doesn't have an user or"
+                        "Given entry doesn't have an user or"
                         " group tag");
         return NULL;
     }
-    return PyInt_FromLong(value);
+    return PyLong_FromUnsignedLong(value);
 }
 
 /* Returns the parent ACL of the entry */
@@ -989,23 +1004,15 @@ static PyObject* Entry_get_parent(PyObject *obj, void* arg) {
  * should be created at init time!
  */
 static PyObject* Entry_get_permset(PyObject *obj, void* arg) {
-    Entry_Object *self = (Entry_Object*)obj;
     PyObject *p;
-    Permset_Object *ps;
 
-    p = Permset_new(&Permset_Type, NULL, NULL);
-    if(p == NULL)
-        return NULL;
-    ps = (Permset_Object*)p;
-    if(acl_get_permset(self->entry, &ps->permset) == -1) {
-        PyErr_SetFromErrno(PyExc_IOError);
-        Py_DECREF(p);
+    PyObject *perm_arglist = Py_BuildValue("(O)", obj);
+    if (perm_arglist == NULL) {
         return NULL;
     }
-    ps->parent_entry = obj;
-    Py_INCREF(obj);
-
-    return (PyObject*)p;
+    p = PyObject_CallObject((PyObject*)&Permset_Type, perm_arglist);
+    Py_DECREF(perm_arglist);
+    return p;
 }
 
 /* Sets the permset of the entry to the passed Permset */
@@ -1013,6 +1020,8 @@ static int Entry_set_permset(PyObject* obj, PyObject* value, void* arg) {
     Entry_Object *self = (Entry_Object*)obj;
     Permset_Object *p;
 
+    ENTRY_SET_CHECK(self, "permset", value);
+
     if(!PyObject_IsInstance(value, (PyObject*)&Permset_Type)) {
         PyErr_SetString(PyExc_TypeError, "argument 1 must be posix1e.Permset");
         return -1;
@@ -1046,8 +1055,7 @@ static PyObject* Entry_copy(PyObject *obj, PyObject *args) {
     if(acl_copy_entry(self->entry, other->entry) == -1)
         return PyErr_SetFromErrno(PyExc_IOError);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 /**** Permset type *****/
@@ -1056,14 +1064,30 @@ static PyObject* Entry_copy(PyObject *obj, PyObject *args) {
 static PyObject* Permset_new(PyTypeObject* type, PyObject* args,
                              PyObject *keywds) {
     PyObject* newpermset;
+    Permset_Object* permset;
+    Entry_Object* parent = NULL;
+
+    if (!PyArg_ParseTuple(args, "O!", &Entry_Type, &parent)) {
+        return NULL;
+    }
 
     newpermset = PyType_GenericNew(type, args, keywds);
 
-    if(newpermset != NULL) {
-        ((Permset_Object*)newpermset)->permset = NULL;
-        ((Permset_Object*)newpermset)->parent_entry = NULL;
+    if(newpermset == NULL) {
+        return NULL;
     }
 
+    permset = (Permset_Object*)newpermset;
+
+    if(acl_get_permset(parent->entry, &permset->permset) == -1) {
+        PyErr_SetFromErrno(PyExc_IOError);
+        Py_DECREF(newpermset);
+        return NULL;
+    }
+
+    permset->parent_entry = (PyObject*)parent;
+    Py_INCREF(parent);
+
     return newpermset;
 }
 
@@ -1075,14 +1099,12 @@ static int Permset_init(PyObject* obj, PyObject* args, PyObject *keywds) {
     if (!PyArg_ParseTuple(args, "O!", &Entry_Type, &parent))
         return -1;
 
-    if(acl_get_permset(parent->entry, &self->permset) == -1) {
-        PyErr_SetFromErrno(PyExc_IOError);
+    if ((PyObject*)parent != self->parent_entry) {
+        PyErr_SetString(PyExc_ValueError,
+                        "Can't reinitialize with a different parent");
         return -1;
     }
 
-    self->parent_entry = (PyObject*)parent;
-    Py_INCREF(parent);
-
     return 0;
 }
 
@@ -1111,7 +1133,7 @@ static PyObject* Permset_str(PyObject *obj) {
     pstr[0] = get_perm(self->permset, ACL_READ) ? 'r' : '-';
     pstr[1] = get_perm(self->permset, ACL_WRITE) ? 'w' : '-';
     pstr[2] = get_perm(self->permset, ACL_EXECUTE) ? 'x' : '-';
-    return MyString_FromStringAndSize(pstr, 3);
+    return PyUnicode_FromStringAndSize(pstr, 3);
 }
 
 static char __Permset_clear_doc__[] =
@@ -1125,9 +1147,7 @@ static PyObject* Permset_clear(PyObject* obj, PyObject* args) {
     if(acl_clear_perms(self->permset) == -1)
         return PyErr_SetFromErrno(PyExc_IOError);
 
-    /* Return the result */
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static PyObject* Permset_get_right(PyObject *obj, void* arg) {
@@ -1145,12 +1165,12 @@ static int Permset_set_right(PyObject* obj, PyObject* value, void* arg) {
     int on;
     int nerr;
 
-    if(!PyInt_Check(value)) {
+    if(!PyLong_Check(value)) {
         PyErr_SetString(PyExc_ValueError, "invalid argument, an integer"
                         " is expected");
         return -1;
     }
-    on = PyInt_AsLong(value);
+    on = PyLong_AsLong(value);
     if(on)
         nerr = acl_add_perm(self->permset, *(acl_perm_t*)arg);
     else
@@ -1186,9 +1206,7 @@ static PyObject* Permset_add(PyObject* obj, PyObject* args) {
     if(acl_add_perm(self->permset, (acl_perm_t) right) == -1)
         return PyErr_SetFromErrno(PyExc_IOError);
 
-    /* Return the result */
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static char __Permset_delete_doc__[] =
@@ -1215,9 +1233,7 @@ static PyObject* Permset_delete(PyObject* obj, PyObject* args) {
     if(acl_delete_perm(self->permset, (acl_perm_t) right) == -1)
         return PyErr_SetFromErrno(PyExc_IOError);
 
-    /* Return the result */
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static char __Permset_test_doc__[] =
@@ -1259,11 +1275,11 @@ static char __ACL_Type_doc__[] =
     "\n"
     ".. note:: only one keyword parameter should be provided\n"
     "\n"
-    ":param string file: creates an ACL representing\n"
-    "    the access ACL of the specified file.\n"
-    ":param string filedef: creates an ACL representing\n"
+    ":param string/bytes/path-like file: creates an ACL representing\n"
+    "    the access ACL of the specified file or directory.\n"
+    ":param string/bytes/path-like filedef: creates an ACL representing\n"
     "    the default ACL of the given directory.\n"
-    ":param int fd: creates an ACL representing\n"
+    ":param int/iostream fd: creates an ACL representing\n"
     "    the access ACL of the given file descriptor.\n"
     ":param string text: creates an ACL from a \n"
     "    textual description; note the ACL must be valid, which\n"
@@ -1274,6 +1290,8 @@ static char __ACL_Type_doc__[] =
     "    (e.g. ``mode=0644``); this is valid only when the C library\n"
     "    provides the ``acl_from_mode call``, and\n"
     "    note that no validation is done on the given value.\n"
+    ":param bytes data: creates an ACL from a serialised form,\n"
+    "    as provided by calling ``__getstate__()`` on an existing ACL\n"
     "\n"
     "If no parameters are passed, an empty ACL will be created; this\n"
     "makes sense only when your OS supports ACL modification\n"
@@ -1291,7 +1309,7 @@ static PyMethodDef ACL_methods[] = {
     {"check", ACL_check, METH_NOARGS, __check_doc__},
     {"equiv_mode", ACL_equiv_mode, METH_NOARGS, __equiv_mode_doc__},
 #endif
-#ifdef HAVE_ACL_COPYEXT
+#ifdef HAVE_ACL_COPY_EXT
     {"__getstate__", ACL_get_state, METH_NOARGS,
      "Dumps the ACL to an external format."},
     {"__setstate__", ACL_set_state, METH_VARARGS,
@@ -1308,12 +1326,7 @@ static PyMethodDef ACL_methods[] = {
 
 /* The definition of the ACL Type */
 static PyTypeObject ACL_Type = {
-#ifdef IS_PY3K
     PyVarObject_HEAD_INIT(NULL, 0)
-#else
-    PyObject_HEAD_INIT(NULL)
-    0,
-#endif
     "posix1e.ACL",
     sizeof(ACL_Object),
     0,
@@ -1321,12 +1334,8 @@ static PyTypeObject ACL_Type = {
     0,                  /* tp_print */
     0,                  /* tp_getattr */
     0,                  /* tp_setattr */
-#ifdef IS_PY3K
     0,                  /* formerly tp_compare, in 3.0 deprecated, in
                            3.5 tp_as_async */
-#else
-    ACL_nocmp,          /* tp_compare */
-#endif
     0,                  /* tp_repr */
     0,                  /* tp_as_number */
     0,                  /* tp_as_sequence */
@@ -1435,12 +1444,7 @@ static char __Entry_Type_doc__[] =
     ;
 /* The definition of the Entry Type */
 static PyTypeObject Entry_Type = {
-#ifdef IS_PY3K
     PyVarObject_HEAD_INIT(NULL, 0)
-#else
-    PyObject_HEAD_INIT(NULL)
-    0,
-#endif
     "posix1e.Entry",
     sizeof(Entry_Object),
     0,
@@ -1547,12 +1551,7 @@ static char __Permset_Type_doc__[] =
 
 /* The definition of the Permset Type */
 static PyTypeObject Permset_Type = {
-#ifdef IS_PY3K
     PyVarObject_HEAD_INIT(NULL, 0)
-#else
-    PyObject_HEAD_INIT(NULL)
-    0,
-#endif
     "posix1e.Permset",
     sizeof(Permset_Object),
     0,
@@ -1616,12 +1615,10 @@ static PyObject* aclmodule_delete_default(PyObject* obj, PyObject* args) {
         return NULL;
 
     if(acl_delete_def_file(filename) == -1) {
-        return PyErr_SetFromErrno(PyExc_IOError);
+        return PyErr_SetFromErrnoWithFilename(PyExc_IOError, filename);
     }
 
-    /* Return the result */
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 #ifdef HAVE_LINUX
@@ -1635,38 +1632,39 @@ static char __has_extended_doc__[] =
 
 /* Check for extended ACL a file or fd */
 static PyObject* aclmodule_has_extended(PyObject* obj, PyObject* args) {
-    PyObject *myarg;
+    PyObject *item, *tmp;
     int nret;
     int fd;
 
-    if (!PyArg_ParseTuple(args, "O", &myarg))
+    if (!PyArg_ParseTuple(args, "O", &item))
         return NULL;
 
-    if(PyBytes_Check(myarg)) {
-        const char *filename = PyBytes_AS_STRING(myarg);
-        nret = acl_extended_file(filename);
-    } else if (PyUnicode_Check(myarg)) {
-        PyObject *o =
-            PyUnicode_AsEncodedString(myarg,
-                                      Py_FileSystemDefaultEncoding, "strict");
-        if (o == NULL)
-            return NULL;
-        const char *filename = PyBytes_AS_STRING(o);
-        nret = acl_extended_file(filename);
-        Py_DECREF(o);
-    } else if((fd = PyObject_AsFileDescriptor(myarg)) != -1) {
-        nret = acl_extended_fd(fd);
+    if((fd = PyObject_AsFileDescriptor(item)) != -1) {
+        if((nret = acl_extended_fd(fd)) == -1) {
+            PyErr_SetFromErrno(PyExc_IOError);
+        }
     } else {
-        PyErr_SetString(PyExc_TypeError, "argument 1 must be string, int,"
-                        " or file-like object");
-        return 0;
-    }
-    if(nret == -1) {
-        return PyErr_SetFromErrno(PyExc_IOError);
+      // PyObject_AsFileDescriptor sets an error when failing, so clear
+      // it such that further code works; some method lookups fail if an
+      // error already occured when called, which breaks at least
+      // PyOS_FSPath (called by FSConverter).
+      PyErr_Clear();
+      if(PyUnicode_FSConverter(item, &tmp)) {
+        char *filename = PyBytes_AS_STRING(tmp);
+        if ((nret = acl_extended_file(filename)) == -1) {
+            PyErr_SetFromErrnoWithFilename(PyExc_IOError, filename);
+        }
+        Py_DECREF(tmp);
+      } else {
+          nret = -1;
+      }
     }
 
-    /* Return the result */
-    return PyBool_FromLong(nret);
+    if (nret < 0) {
+        return NULL;
+    } else {
+        return PyBool_FromLong(nret);
+    }
 }
 #endif
 
@@ -1706,6 +1704,8 @@ static char __posix1e_doc__[] =
     "  - :py:data:`HAS_EXTENDED_CHECK` for the module-level\n"
     "    :py:func:`has_extended` function\n"
     "  - :py:data:`HAS_EQUIV_MODE` for the :py:func:`ACL.equiv_mode` method\n"
+    "  - :py:data:`HAS_COPY_EXT` for the :py:func:`ACL.__getstate__` and\n"
+    "    :py:func:`ACL.__setstate__` functions (pickle protocol)\n"
     "\n"
     "Example:\n"
     "\n"
@@ -1775,10 +1775,11 @@ static char __posix1e_doc__[] =
     ".. py:data:: HAS_EQUIV_MODE\n\n"
     "   denotes support for the equiv_mode function\n"
     "\n"
+    ".. py:data:: HAS_COPY_EXT\n\n"
+    "   denotes support for __getstate__()/__setstate__() on an ACL\n"
+    "\n"
     ;
 
-#ifdef IS_PY3K
-
 static struct PyModuleDef posix1emodule = {
     PyModuleDef_HEAD_INIT,
     "posix1e",
@@ -1787,49 +1788,34 @@ static struct PyModuleDef posix1emodule = {
     aclmodule_methods,
 };
 
-#define INITERROR return NULL
-
 PyMODINIT_FUNC
 PyInit_posix1e(void)
-
-#else
-#define INITERROR return
-
-void initposix1e(void)
-#endif
 {
     PyObject *m, *d;
 
-    Py_TYPE(&ACL_Type) = &PyType_Type;
     if(PyType_Ready(&ACL_Type) < 0)
-        INITERROR;
+        return NULL;
 
 #ifdef HAVE_LEVEL2
-    Py_TYPE(&Entry_Type) = &PyType_Type;
     if(PyType_Ready(&Entry_Type) < 0)
-        INITERROR;
+        return NULL;
 
-    Py_TYPE(&Permset_Type) = &PyType_Type;
     if(PyType_Ready(&Permset_Type) < 0)
-        INITERROR;
+        return NULL;
 #endif
 
-#ifdef IS_PY3K
     m = PyModule_Create(&posix1emodule);
-#else
-    m = Py_InitModule3("posix1e", aclmodule_methods, __posix1e_doc__);
-#endif
     if (m==NULL)
-        INITERROR;
+        return NULL;
 
     d = PyModule_GetDict(m);
     if (d == NULL)
-        INITERROR;
+        return NULL;
 
     Py_INCREF(&ACL_Type);
     if (PyDict_SetItemString(d, "ACL",
                              (PyObject *) &ACL_Type) < 0)
-        INITERROR;
+        return NULL;
 
     /* 23.3.6 acl_type_t values */
     PyModule_AddIntConstant(m, "ACL_TYPE_ACCESS", ACL_TYPE_ACCESS);
@@ -1840,12 +1826,12 @@ void initposix1e(void)
     Py_INCREF(&Entry_Type);
     if (PyDict_SetItemString(d, "Entry",
                              (PyObject *) &Entry_Type) < 0)
-        INITERROR;
+        return NULL;
 
     Py_INCREF(&Permset_Type);
     if (PyDict_SetItemString(d, "Permset",
                              (PyObject *) &Permset_Type) < 0)
-        INITERROR;
+        return NULL;
 
     /* 23.2.2 acl_perm_t values */
     PyModule_AddIntConstant(m, "ACL_READ", ACL_READ);
@@ -1891,7 +1877,12 @@ void initposix1e(void)
     PyModule_AddIntConstant(m, "HAS_EXTENDED_CHECK", LINUX_EXT_VAL);
     PyModule_AddIntConstant(m, "HAS_EQUIV_MODE", LINUX_EXT_VAL);
 
-#ifdef IS_PY3K
-    return m;
+    PyModule_AddIntConstant(m, "HAS_COPY_EXT",
+#ifdef HAVE_ACL_COPY_EXT
+                            1
+#else
+                            0
 #endif
+                            );
+    return m;
 }
